#|
This file is part of Ordo.

Copyright (C) 2025 Ray Miller

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, version 3.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.
|#

(define-module (ordo inventory)
  #:use-module (ice-9 eval-string)
  #:use-module (ice-9 match)
  #:use-module (ice-9 textual-ports)
  #:use-module (oop goops)
  #:use-module ((ordo connection) #:select (local-connection))
  #:use-module (ordo logger)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-9)
  #:use-module (srfi srfi-69)
  #:export (<host>
            host
            host?
            host-name
            host-connection
            host-tags

            <host-resolver>

            <inventorp>
            get-inventory-hosts
            resolve-hosts
            load-inventory))

(define-class <host> ()
  (name #:init-keyword #:name #:getter host-name)
  (connection #:init-keyword #:connection #:getter host-connection)
  (tags #:init-keyword #:tags #:getter host-tags))

(define (host? h) (is-a? h <host>))

(define (host . args) (apply make <host> args))

(define (tagged-every? wanted-tags)
  (lambda (h)
    (lset= equal? wanted-tags (lset-intersection equal? (host-tags h) wanted-tags))))

(define (tagged-any? wanted-tags)
  (lambda (h)
    (not (null? (lset-intersection equal? (host-tags h) wanted-tags)))))

(define (named? hostname)
  (lambda (h)
    (string=? (host-name h) hostname)))

(define-class <host-resolver> ())

(define-class <inventory> (<host-resolver>)
  (hosts #:init-keyword #:hosts #:getter get-inventory-hosts))

(define-method (resolve-hosts (inventory <inventory>) expr)
  (define hosts (get-inventory-hosts inventory))
  (match expr
    ("localhost" (list (or (find (named? "localhost") hosts)
                           (host #:name "localhost" #:connection (local-connection)))))
    ((? string? hostname) (filter (named? hostname) hosts))
    ('all hosts)
    (('tagged tag) (filter (tagged-every? (list tag)) hosts))
    (('tagged/every tag . tags) (filter (tagged-every? (cons tag tags)) hosts))
    (('tagged/any tag . tags) (filter (tagged-any? (cons tag tags)) hosts))
    ((. hostnames) (filter (lambda (h) (member (host-name h) hostnames string=?)) hosts))))

(define (load-inventory filename)
  (log-msg 'INFO "Loading inventory " filename)
  (let* ((hosts (eval-string (call-with-input-file filename get-string-all)
                             #:file filename))
         (hosts (if (list? hosts) hosts '())))
    (when (null? hosts)
      (log-msg 'NOTICE "Inventory is empty, only localhost will be available"))
    (make <inventory> #:hosts hosts)))
